import json
import numpy as np
import pandas as pd
from pykrige import OrdinaryKriging
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import StandardScaler

from landsklim.lisdqs.lisdqs_analysis import LISDQSAnalysis, LISDQSAnalysisMode
from landsklim.lisdqs.lisdqs_configuration import LISDQSConfiguration
from landsklim.lisdqs.lisdqs_interpolation import LISDQSInterpolation, LISDQSInterpolationType
from landsklim.lisdqs.lisdqs_project import LISDQSProject
from landsklim.lisdqs.map_layer import MapLayer
from landsklim.lisdqs.phase import IPhase
from landsklim.lisdqs.polygons_definition import PolygonsDefinition
from landsklim.lisdqs.regression_model import MultipleRegressionModel, SmoothingMode
from landsklim.lisdqs.regressor import Regressor


PUBLIC_ENUMS = {
    'LISDQSAnalysisMode': LISDQSAnalysisMode,
    'SmoothingMode': SmoothingMode,
    'LISDQSInterpolationType': LISDQSInterpolationType
}


class LISDQSEncoder(json.JSONEncoder):
    def default(self, obj):

        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)

        if isinstance(obj, LinearRegression):
            return obj.__dict__
        if isinstance(obj, StandardScaler):
            return obj.__dict__
        if isinstance(obj, OrdinaryKriging):
            return str(obj)

        if isinstance(obj, LISDQSProject):
            return obj.to_json()
        if isinstance(obj, LISDQSConfiguration):
            return obj.to_json()
        if isinstance(obj, LISDQSAnalysis):
            return obj.to_json()
        if isinstance(obj, MultipleRegressionModel):
            return obj.to_json()
        if isinstance(obj, LISDQSInterpolation):
            return obj.to_json()
        if isinstance(obj, IPhase):
            return obj.to_json()
        if isinstance(obj, Regressor):
            return obj.to_json()
        if isinstance(obj, MapLayer):
            return obj.to_json()
        if isinstance(obj, PolygonsDefinition):
            return obj.to_json()
        if type(obj) in PUBLIC_ENUMS.values():
            return {"__enum__": str(obj)}
        if isinstance(obj, pd.DataFrame):
            return obj.to_dict()
        if isinstance(obj, np.ndarray):
            return [obj.tolist(), str(obj.dtype)]
        return super().default(obj)

    # Decoder : reconstruire les references hiérarchiques
    # Decoder : les enums
    # Decoder : LISDQSAnalysis._pearson_correlation
    # Decoder : Regressor et Phase : héritage
    # Decoder : reconstruire les tableaux Numpy et les DataFrame
        # if isinstance(obj, np.ndarray):
        #   np.asarray(obj[0], dtype=obj[1])
    # Decoder : reconstruire les modèles de regression et les scalers
    # Decoder : reconstruire les modèles de krigeage
