from typing import List, Optional

from landsklim.lisdqs.phase import IPhase
from landsklim.lisdqs.phase_polynomial import PhasePolynomial
from landsklim.lisdqs.phase_multiple_regression import PhaseMultipleRegression
from landsklim.lisdqs.phase_kriging import PhaseKriging
from landsklim.lisdqs.phase_composite import PhaseComposite


class PhaseFactory:

    @staticmethod
    def get_phase(phase_name: str, **kwargs) -> IPhase:
        registered_classes: List[type(IPhase)] = [PhaseMultipleRegression, PhasePolynomial, PhaseKriging, PhaseComposite]
        phase: Optional[IPhase] = None
        for phase_class in registered_classes:  # type: IPhase
            if phase_name == phase_class.class_name():
                phase = phase_class(**kwargs)
        return phase
