from typing import List
import os

from qgis.core import QgsRasterLayer, QgsProject
from qgis import processing

from landsklim.lisdqs.regressor import Regressor
from landsklim.lisdqs.landsklim_constants import LAYER_TYPE_PATH, LandsklimLayerType
from landsklim.lisdqs.regressor_slope import RegressorSlope
from landsklim.processing.algorithm_orientation import OrientationProcessingAlgorithm


class RegressorOrientation(Regressor):

    def __init__(self, windows: int, polynomial_degree: int):
        super().__init__(windows, polynomial_degree, ["local_min", "local_max", "a0", "a1"])

    @staticmethod
    def class_name() -> str:
        return "REGRESSOR_ORIENTATION"

    @staticmethod
    def min_window() -> int:
        return 3

    def prefix(self) -> str:
        return "orien"

    def name(self) -> str:
        return Regressor.tr("Orientation")

    def compute(self, source_raster: QgsRasterLayer):
        raster_path, exists = self._create_path()

        project_path = self.project_path()

        if not exists:
            from landsklim.lisdqs.regressor_factory import RegressorFactory
            source_slope = RegressorFactory.get_regressor(RegressorSlope.class_name(), self.get_windows(), 1).get_path()
            params = {
                'INPUT': source_raster,
                OrientationProcessingAlgorithm.INPUT_SLOPE: source_slope,
                'INPUT_WINDOW': self.get_windows(),
                'INPUT_CUSTOM_NO_DATA': None,
                'ADDITIONAL_VARIABLES_FOLDER': os.path.join(project_path, LAYER_TYPE_PATH[LandsklimLayerType.AdditionalLayer]),
                'OUTPUT': raster_path
            }
            processing.run("landsklim:orientation", params)
